# Generated by Django 5.0.6 on 2024-06-18 08:48

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0061_portfolio_bank_accounts_product_bank_account_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="dailyportfoliocashflow",
            options={
                "permissions": [("administrate_dailyportfoliocashflow", "Can administrate Daily Portfolio CashFlow")],
                "verbose_name": "Daily Portfolio CashFlow",
                "verbose_name_plural": "Daily Portfolio CashFlow",
            },
        ),
    ]
